@@ -23,6 +23,7 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 """
 
+import json
 import socket
 import os
 import re
@@ -38,43 +39,35 @@ from typing import Callable, Dict, List, Tuple
 
 from simple_http_server import ControllerFunction, StaticFile
 
-from .base_http_request_handler import BaseHTTPRequestHandler
-from .__utils import remove_url_first_slash
+from .base_request_handler import BaseHTTPRequestHandler
+from .wsgi_request_handler import WSGIRequestHandler
+from .__utils import remove_url_first_slash, get_function_args, get_function_kwargs
 from .logger import get_logger
 
 _logger = get_logger("simple_http_server.http_server")
 
 
-class HTTPServer(ThreadingMixIn, TCPServer):
-
-    allow_reuse_address = 1    # Seems to make sense in testing environment
+class RoutingConf:
 
     HTTP_METHODS = ["OPTIONS", "GET", "HEAD", "POST", "PUT", "DELETE", "TRACE", "CONNECT"]
 
-    def server_bind(self):
-        """Override server_bind to store the server name."""
-        TCPServer.server_bind(self)
-        host, port = self.server_address[:2]
-        self.server_name = socket.getfqdn(host)
-        self.server_port = port
-
-    def __init__(self, addr, res_conf={}):
-        super().__init__(addr, BaseHTTPRequestHandler)
+    def __init__(self, res_conf={}):
         self.method_url_mapping: Dict[str, Dict[str, ControllerFunction]] = {"_": {}}
         self.path_val_url_mapping: Dict[str, Dict[str, ControllerFunction]] = {"_": OrderedDict()}
         self.method_regexp_mapping: Dict[str, Dict[str, ControllerFunction]] = {"_": OrderedDict()}
-        for mth in HTTPServer.HTTP_METHODS:
+        for mth in self.HTTP_METHODS:
             self.method_url_mapping[mth] = {}
             self.path_val_url_mapping[mth] = OrderedDict()
             self.method_regexp_mapping[mth] = OrderedDict()
 
         self.filter_mapping = OrderedDict()
         self._res_conf = []
+        self.add_res_conf(res_conf)
 
         self.ws_mapping = OrderedDict()
         self.ws_path_val_mapping = OrderedDict()
 
-        self.add_res_conf(res_conf)
+        self.error_page_mapping = {}
 
     @property
     def res_conf(self):
@@ -130,7 +123,7 @@ class HTTPServer(ThreadingMixIn, TCPServer):
         regexp = ctrl.regexp
         method = ctrl.method
         _logger.debug(f"map url {url}|{regexp} with method[{method}] to function {ctrl.func}. ")
-        assert method is None or method == "" or method.upper() in HTTPServer.HTTP_METHODS
+        assert method is None or method == "" or method.upper() in self.HTTP_METHODS
         _method = method.upper() if method is not None and method != "" else "_"
         if regexp:
             self.method_regexp_mapping[_method][regexp] = ctrl
@@ -232,7 +225,7 @@ class HTTPServer(ThreadingMixIn, TCPServer):
                 available_filters.append(val)
         return available_filters
 
-    def map_websocket_hanlder(self, endpoint, handler_class):
+    def map_websocket_handler(self, endpoint, handler_class):
         url = remove_url_first_slash(endpoint)
         path_pattern, path_names = self.__get_path_reg_pattern(url)
         if path_pattern is None:
@@ -258,6 +251,107 @@ class HTTPServer(ThreadingMixIn, TCPServer):
                 return clz, path_values
         return None, {}
 
+    def map_error_page(self, code: str, error_page_fun: Callable):
+        if not code:
+            c = "_"
+        else:
+            c = str(code).lower()
+        self.error_page_mapping[c] = error_page_fun
+
+    def _default_error_page(self, code: int, message: str = "", explain: str = ""):
+        return json.dumps({
+            "code": code,
+            "message": message,
+            "explain": explain
+        })
+
+    def error_page(self, code: int, message: str = "", explain: str = ""):
+        c = str(code)
+        func = None
+        if c in self.error_page_mapping:
+            func = self.error_page_mapping[c]
+        elif code > 200:
+            c0x = c[0:2] + "x"
+            if c0x in self.error_page_mapping:
+                func = self.error_page_mapping[c0x]
+            elif "_" in self.error_page_mapping:
+                func = self.error_page_mapping["_"]
+
+        if not func:
+            func = self._default_error_page
+        _logger.debug(f"error page function:: {func}")
+
+        co = code
+        msg = message
+        exp = explain
+
+        args_def = get_function_args(func, None)
+        kwargs_def = get_function_kwargs(func, None)
+
+        args = []
+        for n, t in args_def:
+            _logger.debug(f"set value to error_page function -> {n}")
+            if co is not None:
+                if t is None or t == int:
+                    args.append(co)
+                    co = None
+                    continue
+            if msg is not None:
+                if t is None or t == str:
+                    args.append(msg)
+                    msg = None
+                    continue
+            if exp is not None:
+                if t is None or t == str:
+                    args.append(exp)
+                    exp = None
+                    continue
+            args.append(None)
+
+        kwargs = {}
+        for n, v, t in kwargs_def:
+            if co is not None:
+                if (t is None and isinstance(v, int)) or t == int:
+                    kwargs[n] = co
+                    co = None
+                    continue
+            if msg is not None:
+                if (t is None and isinstance(v, str)) or t == str:
+                    kwargs[n] = msg
+                    msg = None
+                    continue
+            if exp is not None:
+                if (t is None and isinstance(v, str)) or t == str:
+                    kwargs[n] = exp
+                    exp = None
+                    continue
+            kwargs[n] = v
+
+        if args and kwargs:
+            return func(*args, **kwargs)
+        elif args:
+            return func(*args)
+        elif kwargs:
+            return func(**kwargs)
+        else:
+            return func()
+
+
+class HTTPServer(ThreadingMixIn, TCPServer, RoutingConf):
+
+    allow_reuse_address = 1    # Seems to make sense in testing environment
+
+    def server_bind(self):
+        """Override server_bind to store the server name."""
+        TCPServer.server_bind(self)
+        host, port = self.server_address[:2]
+        self.server_name = socket.getfqdn(host)
+        self.server_port = port
+
+    def __init__(self, addr, res_conf={}):
+        TCPServer.__init__(self, addr, BaseHTTPRequestHandler)
+        RoutingConf.__init__(self, res_conf)
+
 
 class SimpleDispatcherHttpServer:
     """Dispatcher Http server"""
@@ -265,11 +359,14 @@ class SimpleDispatcherHttpServer:
     def map_filter(self, path_pattern, filter_fun):
         self.server.map_filter(path_pattern, filter_fun)
 
-    def map_request(self, ctrl: ControllerFunction):
+    def map_controller(self, ctrl: ControllerFunction):
         self.server.map_controller(ctrl)
 
     def map_websocket_handler(self, endpoint, handler_class):
-        self.server.map_websocket_hanlder(endpoint, handler_class)
+        self.server.map_websocket_handler(endpoint, handler_class)
+
+    def map_error_page(self, code, func):
+        self.server.map_error_page(code, func)
 
     def __init__(self,
                  host: Tuple[str, int] = ('', 9090),
@@ -282,6 +379,7 @@ class SimpleDispatcherHttpServer:
                  ssl_context: _ssl.SSLContext = None,
                  resources: Dict[str, str] = {}):
         self.host = host
+        self.__ready = False
 
         self.ssl = ssl
         self.server = HTTPServer(self.host, res_conf=resources)
@@ -299,6 +397,10 @@ class SimpleDispatcherHttpServer:
                 server_side=True
             )
 
+    @property
+    def ready(self):
+        return self.__ready
+
     def resources(self, res={}):
         self.server.res_conf = res
 
@@ -308,10 +410,23 @@ class SimpleDispatcherHttpServer:
         else:
             ssl_hint = ""
         _logger.info(f"Dispatcher Http Server starts. Listen to port [{self.host[1]}]{ssl_hint}.")
-        self.server.serve_forever()
+        try:
+            self.__ready = True
+            self.server.serve_forever()
+        except:
+            self.__ready = False
+            raise
 
     def shutdown(self):
         # server must shutdown in a separate thread, or it will be deadlocking...WTF!
-        t = threading.Thread(target=self.server.shutdown)
-        t.daemon = True
-        t.start()
+        threading.Thread(target=self.server.shutdown, daemon=True).start()
+
+
+class WSGIProxy(RoutingConf):
+
+    def __init__(self, res_conf):
+        super().__init__(res_conf=res_conf)
+
+    def app_proxy(self, environment, start_response):
+        requestHandler = WSGIRequestHandler(self, environment, start_response)
+        return requestHandler.handle()
